import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class BytePairEncoding {

	/** 置換情報を保持しておくクラス */
	public static class Convert {
		final public String original;
		final public char replace;

		Convert(final String original, final char replace) {
			this.original = original;
			this.replace = replace;
		}

		@Override
		public String toString() {
			return this.original;
		}
	}

	/** 圧縮結果を表すクラス */
	public static class Compressed {
		final public List<Convert> convertList;
		final public String output;

		Compressed(final List<Convert> convertList, final String output) {
			this.convertList = convertList;
			this.output = output;
		}
	}

	private char getUnusedChar(final String base) {
		for (char ch = Character.MIN_CODE_POINT; ch < Character.MAX_CODE_POINT; ++ch) {
			if (Character.isDefined(ch) && !Character.isHighSurrogate(ch)
					&& !Character.isLowSurrogate(ch)) {
				if (base.indexOf(ch) < 0) {
					return ch;
				}
			}
		}
		return 0;
	}

	private Compressed compress(String text) {
		Map<String, Integer> freqMap = new HashMap<String, Integer>();
		List<Convert> convertList = new ArrayList<Convert>();
		String targetPair;

		for (;;) {
			// Convertをファイルに書き出すのに3文字分使用するため、実際に置換するのは3回より大きいペアのみとする。
			int maxFreq = 3;
			targetPair = null;
			String prevPair = null;
			// 素直に数え上げを行う
			for (int cursor = 0; cursor < text.length() - 1; ++cursor) {
				String pair = text.substring(cursor, cursor + 2);
				// 同じ文字が3つ以上続いているとき、重複してカウントしないようにする
				if (pair.equals(prevPair)) {
					prevPair = null;
					continue;
				}
				prevPair = pair;

				Integer freq = freqMap.get(pair);
				if (freq == null) {
					freq = 0;
				}
				freqMap.put(pair, freq + 1);
				// 4回以上使われている、最多の2文字を記録する
				if (maxFreq < freq + 1) {
					maxFreq = freq + 1;
					targetPair = pair;
				}
			}
			if (targetPair == null) {
				break;
			}
			// 使用されていないcharを取得する
			char replace = getUnusedChar(text);
			// 置換する
			text = text.replace(targetPair, String.valueOf(replace));
			// 置換したことを順序と共に記録しておく
			convertList.add(new Convert(targetPair, replace));
			freqMap.clear();
		}
		return new Compressed(convertList, text);
	}

	private String uncompress(final Compressed compressed) {
		String ret = compressed.output;
		// 圧縮の逆順に置換する
		for (int i = compressed.convertList.size() - 1; 0 <= i; --i) {
			Convert conv = compressed.convertList.get(i);
			ret = ret.replace(Character.toString(conv.replace), conv.original);
		}
		return ret;
	}

	private static void usageAndExit() {
		System.err.println("usage> BytePairEncoding (target-file)");
		System.err.println("# target-file must exist and contain UTF-8 text.");
		System.exit(1);
	}

	private static File checkOutput(final String output) throws IOException {
		File ret = new File(output);
		if (!ret.createNewFile()) {
			System.err.println("output-file already exists!");
			System.exit(2);
		}
		return ret;
	}

	public static void main(final String[] args) throws IOException {
		BytePairEncoding self = new BytePairEncoding();

		if (args.length != 1) {
			usageAndExit();
		}
		File file = new File(args[0]);
		if (!file.isFile()) {
			usageAndExit();
		}

		BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(file),
				"UTF-8"));
		if (file.getName().endsWith(".bpe")) {
			List<Convert> convertList = new ArrayList<Convert>();
			for (int count = Integer.parseInt(reader.readLine()); 0 < count; --count) {
				char[] buff = new char[3];
				reader.read(buff);
				convertList.add(new Convert(String.valueOf(buff, 1, 2), buff[0]));
			}
			StringBuffer contentBuffer = new StringBuffer();
			for (int ch = reader.read(); ch != -1; ch = reader.read()) {
				contentBuffer.append((char) ch);
			}
			String original = self
					.uncompress(new Compressed(convertList, contentBuffer.toString()));
			BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(
					checkOutput(file.getPath().replaceAll(".bpe$", ""))), "UTF-8"));
			writer.append(original);
			writer.flush();
			writer.close();
		} else {
			StringBuffer contentBuffer = new StringBuffer();
			for (int ch = reader.read(); ch != -1; ch = reader.read()) {
				contentBuffer.append((char) ch);
			}
			reader.close();

			Compressed compressed = self.compress(contentBuffer.toString());

			BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(
					checkOutput(file.getPath() + ".bpe")), "UTF-8"));
			writer.append(Integer.toString(compressed.convertList.size())).append('\n');
			for (Convert conv : compressed.convertList) {
				writer.append(conv.replace).append(conv.original);
			}
			writer.append(compressed.output);
			writer.flush();
			writer.close();
		}
	}
}
