import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;

/**
 * 辞書圧縮のサンプル
 * 
 */
public class Dictionary {
	/** 圧縮においてスライド辞書方式(FastSlideDictionary.java)を使うかどうか */
	private static final boolean USING_SLIDE = false;

	/** inputの中のleftCursor部分とrightCursor部分が何文字一致しているかを返す */
	protected int calcMatchingLength(final String input, final int leftCursor, final int rightCursor) {
		assert leftCursor < rightCursor;
		for (int length = 0; rightCursor + length < input.length(); ++length) {
			if (input.charAt(leftCursor + length) != input.charAt(rightCursor + length)) {
				return length;
			}
		}
		return 0;
	}

	public String compress(final String input) {
		StringBuffer buffer = new StringBuffer();

		for (int cursor = 0; cursor < input.length();) {
			int distance = 0;
			int maxLength = 0;
			// 最長一致部分の探索
			for (int distanceCandidate = 0; distanceCandidate < cursor; ++distanceCandidate) {
				int length = calcMatchingLength(input, cursor - distanceCandidate, cursor);
				// 一致長が最長なら保存しておく
				if (maxLength < length) {
					maxLength = length;
					distance = distanceCandidate;
				}
			}
			String dl = String.format("[%d,%d]", distance, maxLength);
			if (dl.length() < maxLength) {
				// distanceとmaxLengthの組をファイルに出力する（圧縮）
				buffer.append(dl);
				cursor += maxLength;
			} else {
				// input.charAt(cursor)をファイルに出力する（無圧縮）
				char ch = input.charAt(cursor);
				if (ch == '[') {
					buffer.append("[[");
				} else {
					buffer.append(ch);
				}
				cursor += 1;
			}
		}
		return buffer.toString();
	}

	public String uncompress(final String input) {
		StringBuffer buffer = new StringBuffer();

		for (int cursor = 0; cursor < input.length();) {
			char ch = input.charAt(cursor);
			if (ch != '[') {
				buffer.append(ch);
				cursor += 1;
			} else if (input.charAt(ch + 1) == '[') {
				buffer.append('[');
				cursor += 2;
			} else {
				int comma = input.indexOf(',', cursor + 2);
				int end = input.indexOf(']', comma + 2);

				int distance = Integer.parseInt(input.substring(cursor + 1, comma));
				int length = Integer.parseInt(input.substring(comma + 1, end));

				// StringBuffer内でコピー元、コピー先が被ったときのためにループ処理
				for (int copy = Math.min(distance, length); 0 < length; copy = Math.min(distance,
						length)) {
					buffer.append(buffer, buffer.length() - distance, buffer.length() - distance
							+ copy);
					length -= copy;
				}
				cursor = end + 1;
			}
		}
		return buffer.toString();
	}

	private static void usageAndExit() {
		System.err.println("usage> Dictionary (target-file)");
		System.err.println("# target-file must exist and contain UTF-8 text.");
		System.exit(1);
	}

	private static File checkOutput(final String output) throws IOException {
		File ret = new File(output);
		if (!ret.createNewFile()) {
			System.err.println("output-file already exists!");
			System.exit(2);
		}
		return ret;
	}

	public static void main(final String[] args) throws IOException {
		Dictionary self;
		if (USING_SLIDE) {
			self = new FastSlideDictionary();
		} else {
			self = new Dictionary();
		}

		if (args.length != 1) {
			usageAndExit();
		}
		File file = new File(args[0]);
		if (!file.isFile()) {
			usageAndExit();
		}

		BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(file),
				"UTF-8"));
		if (file.getName().endsWith(".sd")) {
			StringBuffer contentBuffer = new StringBuffer();
			for (int ch = reader.read(); ch != -1; ch = reader.read()) {
				contentBuffer.append((char) ch);
			}
			String original = self.uncompress(contentBuffer.toString());
			BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(
					checkOutput(file.getPath().replaceAll(".sd$", ""))), "UTF-8"));
			writer.append(original);
			writer.flush();
			writer.close();
		} else {
			StringBuffer contentBuffer = new StringBuffer();
			for (int ch = reader.read(); ch != -1; ch = reader.read()) {
				contentBuffer.append((char) ch);
			}
			reader.close();

			String result = self.compress(contentBuffer.toString());

			BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(
					checkOutput(file.getPath() + ".sd")), "UTF-8"));
			writer.append(result);
			writer.flush();
			writer.close();
		}
	}
}
